# -*- coding: utf-8 -*-
from copy import copy
from mediatumfsm.fsm import FSM


class SessionFSM(FSM):
    """FSM which saves its current state to a Session (from Flask, for example)"""
    # fields which will be ignored when saving current state data
    # can be used for transient caching fields etc.
    ignored_data_fields = None
    # must be set by subclass to an object like werkzeug.LocalProxy which proxies a dict-like object
    session = None

    def save_state(self):
        """Saves the current state and associated state data.
        Fields in self.ignored_data_fields are skipped.
        """
        if self.name is not None:
            self.session.setdefault(self.name, {})
            o = self.session[self.name]
        else:
            o = self.session
        o["current_state"] = self.state.__name__
        if self.ignored_data_fields:
            state_data = copy(self.data)
            for field in self.ignored_data_fields:
                try:
                    del state_data.__dict__[field]
                except KeyError:
                    pass
                except AttributeError:
                    break
            o["state_data"] = state_data
        else:
            o["state_data"] = self.data

    @classmethod
    def from_saved_state(cls, name=None):
        """Restore FSM object from saved data
        :param name: name(space) of the FSM, needed when multiple FSM should coexist in one session
        """
        if name is not None:
            o = cls.session.get(name, {})
        else:
            o = cls.session
        return cls(o.get("current_state"), o.get("state_data"), name=name)

    def clear(self):
        """Removes saved state from session"""
        if self.name is not None:
            self.session.setdefault(self.name, {})
            o = self.session[self.name]
        else:
            o = self.session
        if "current_state" in o:
            del o["current_state"]
        if "state_data" in o:
            del o["state_data"]


    def reset(self):
        """Clears session state and calls FSM reset"""
        self.clear()
        super(SessionFSM, self).reset()



# -*- coding: utf-8 -*-
from mediatumfsm.fsm import FSM, State, case, Goto, unhandled_case

class Message(object):
    pass

class SwitchOn(Message):
    pass

class SwitchOff(Message):
    pass

class Smash(Message):
    pass


electricity = False


class LightBulb(FSM):
    initial_state = "Off"

    class Off(State):
        @case(SwitchOn, electricity)
        def on(*_):
            return Goto(LightBulb.On)

    class On(State):
        @case(SwitchOff)
        def off(*_):
            return Goto(LightBulb.Off)

    class Broken(State):
        pass

    @unhandled_case(Smash)
    def break_it(m, d):
        return Goto(LightBulb.Broken), d



# -*- coding: utf-8 -*-
from __future__ import print_function
import logging
import time
from mediatumfsm.fsm import State, case, unhandled_case, on_transition, Stay, Goto
from mediatumfsm.sessionfsm import SessionFSM

logg = logging.getLogger("fsm")

def even(m, _):
    return m % 2 == 0

transition_log = {}

switched_to_int_times = 0

class Message(object):
    pass


class AcceptInt(Message):
    pass


class AcceptStr(Message):
    pass


class Accumulator(SessionFSM):
    initial_data = False
    initial_state_name = "Nothing"
    session = {}

    class Nothing(State):
        pass

    class Str(State):
        @case(AcceptStr)
        def nothing(*_):
            return Stay

        @case(str)
        def handle_str(message, data):
            print("string message:", message, "data", data)
            return Stay, data + message

    class Int(State):
        @case(AcceptInt)
        def nothing(*_):
            return Stay

        @case(int, even)
        def set_value_even_int(number, data):
            print("got even number")
            return Stay, (data[0] + number, "even")

        @case(int, lambda m, d: m % 2 == 1)
        def set_value_odd_int(number, data):
            print("got odd number")
            return Stay, (data[0] + number, "odd")

    # switching to Str and Int state is always allowed

    @unhandled_case(AcceptInt)
    def switch_to_int(*_):
        return Goto(Accumulator.Int), (0, None)

    @unhandled_case(AcceptStr)
    def switch_to_str(*_):
        return Goto(Accumulator.Str), ""

    @unhandled_case(int)
    def invalid_int(*_):
        raise ValueError("int not allowed here!")

    @on_transition(State, "Int")
    def switched_to_int(*_):
        global switched_to_int_times
        switched_to_int_times += 1
        print("switched to int accepting state!")

    @on_transition(State, State)
    def trace(msg, data, src_state, dest_state):
        transition_log[time.time()] = {"msg": msg, "data": data, "src_state": src_state, "dest_state": dest_state}
        logg.info("state transition from %s to %s on msg %s with data '%s'", src_state.name, dest_state.name, msg, data)



# -*- coding: utf-8 -*-
import ast
from copy import copy
from inspect import isclass
import inspect
import logging
from mediatumfsm.compat import itervalues, iteritems, iterkeys
logg = logging.getLogger("fsm")


def _sort_handlers(handlers):
    handlers.sort(key=lambda h: h.__code__.co_firstlineno)
    handlers.sort(key=lambda h: 0 if h._condition else 1)
    handlers.sort(key=lambda h: h._pos)


def _check_handler_validity(handlers, msg_type, func):
    if msg_type in handlers:
        # check for duplicate handlers with same condition
        for h in handlers[msg_type]:
            if h._condition == func._condition:
                logg.warn("handler (without condition) for msg type %s already defined, ignoring it: %s",
                          msg_type, func)
                return False
    return True


class _StateMeta(type):
    def __init__(cls, name, bases, dct):
        handlers = {}
        for name, obj in iteritems(dct):
            if isinstance(obj, staticmethod):
                func = obj.__func__
                msg_type = getattr(func, "_message_type", None)
                if msg_type:
                    valid = _check_handler_validity(handlers, msg_type, func)
                    if valid:
                        h_for_msg_type = handlers.setdefault(msg_type, [])
                        h_for_msg_type.append(func)
        for h_for_msg_type in itervalues(handlers):
            _sort_handlers(h_for_msg_type)
        cls.handlers = handlers
        super(_StateMeta, cls).__init__(name, bases, dct)


class _FSMMeta(type):
    def __init__(cls, name, bases, dct):
        # get states from bases
        states = {}
        for base in bases:
            if type(base) == _FSMMeta:
                states.update(base.states)
        handlers = {}
        transition_actions = {}
        cls_module = inspect.getmodule(cls)
        for name, obj in iteritems(dct):
            if isclass(obj) and type(obj) == _StateMeta:
                obj.name = obj.__name__
                states[name] = obj
                # register state classes in module where FSM was defined
                setattr(cls_module, name, obj)
                if obj.__name__ == cls.initial_state_name:
                    cls.initial_state = obj
            elif isinstance(obj, staticmethod):
                func = obj.__func__
                # only transition handlers have a _src_state attrib
                src_state = getattr(func, "_src_state", None)
                if src_state:
                    dest_state = func._dest_state
                    transition_actions[(src_state, dest_state)] = func
                # handlers must be defined with @unhandled_case (_bound_to_state == False) here
                bound_to_state = getattr(func, "_bound_to_state", False)
                if bound_to_state:
                    raise Exception("case handler not allowed here, must be defined in state class!")
                msg_type = getattr(func, "_message_type", None)
                if msg_type:
                    h_for_msg_type = handlers.setdefault(msg_type, [])
                    h_for_msg_type.append(func)
        cls.states = states
        cls.state_names = list(iterkeys(states))
        for h_for_msg_type in itervalues(handlers):
            _sort_handlers(h_for_msg_type)
        cls.handlers = handlers
        cls.transition_actions = transition_actions
        super(_FSMMeta, cls).__init__(name, bases, dct)


### state changes

class _StateChange(object):
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from pytest import fixture
from mediatumfsm.test.accumulatorfsm import Accumulator, AcceptStr, Str

### fixtures

@fixture
def session():
    s = Accumulator.session
    s.clear()
    return s

@fixture
def fsm():
    return Accumulator()

@fixture
def fsm_with_name():
    return Accumulator(name="test")


### tests

def test_state_saving(fsm, session):
    fsm.send(AcceptStr())
    assert session["state_data"] == ""
    assert session["current_state"] == "Str"
    fsm.send("spam")
    assert session["state_data"] == "spam"
    new_fsm = Accumulator.from_saved_state()
    assert new_fsm.data == "spam"
    assert new_fsm.state == Str


def test_state_saving_named_fsm(fsm_with_name, session):
    fsm_with_name.send(AcceptStr())
    assert session["test"]["state_data"] == ""
    assert session["test"]["current_state"] == "Str"
    fsm_with_name.send("spam")
    assert session["test"]["state_data"] == "spam"
    new_fsm = Accumulator.from_saved_state("test")
    # compares name, state and data equality
    assert new_fsm == fsm_with_name



# -*- coding: utf-8 -*-
from mediatumfsm.fsm import FSM, State, case, Goto, Stay, Stop

### predefined handler conditions

is_a = lambda m, _: m == "a"
is_fourth_a = lambda m, d: m == "a" and d == 3


class RegexFSM(FSM):
    """Parses the regex 'aaaa(.*)#' and replace upper letters by a star'
    """
    initial_data = 0
    initial_state_name = "A"

    ### states

    class A(State):
        """Skip exactly 4 a's"""
        @case(str, is_fourth_a)
        def finish_a(*_):
            return Goto(RegexFSM.Repeat), ""

        @case(str, is_a)
        def collect_a(msg, data):
            return Stay, data + 1

    class Repeat(State):
        """repeat until # is reached"""
        @case(str, lambda m, _: m == "#")
        def end(*_):
            return Stop

        @case(str)
        def any(m, d):
            if m.isupper():
                m = "*"
            return Goto(RegexFSM.Repeat), d + m

    ### some parsing helpers

    def parse(self, data):
        """Sends all chars of string `data` into the FSM.
        Returns the recognized string if no errors occur.
        :param data: string to parse
        :type data: str
        """
        for c in data:
            self.send(c)
            if self.stopped:
                return self.data

    def parse_gen(self, data):
        """Generator which skips the leading a's and returns the recognized string after each char.
        :param data: string to parse
        :type data: str
        """
        it = iter(data)
        while self.state != RegexFSM.Repeat:
            self.send(next(it))
        while not self.stopped:
            yield self.data
            self.send(next(it))



from setuptools import setup
import codecs
import os
import re

here = os.path.abspath(os.path.dirname(__file__))

# Read the version number from a source file.
# Why read it, and not import?
# see https://groups.google.com/d/topic/pypa-dev/0PkjVpcxTzQ/discussion
def find_version(*file_paths):
    # Open in Latin-1 so that we avoid encoding errors.
    # Use codecs.open for Python 2 compatibility
    with codecs.open(os.path.join(here, *file_paths), 'r', 'latin1') as f:
        version_file = f.read()

    # The version line must have the form
    # __version__ = 'ver'
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="mediatumfsm",
    version=find_version('mediatumfsm', '__init__.py'),
    description="declarative finite state machine library inspired by Akka FSM and erlang gen_fsm",
    # The project URL.
    url='https://mediatumdev.ub.tum.de/projects/mediatum-fsm',
    # Author details
    author='Tobias Stenzel',
    author_email='tobias.stenzel@tum.de',
    # Choose your license
    license='GNU General Public License v3 (GPLv3)',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    keywords='state machine fsm declarative dsl',
    packages=["mediatumfsm"],
    install_requires=["pydot2"],
    setup_requires=["setuptools-git"],
    include_package_data=True,
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function
from pytest import fixture, raises
from mediatumfsm.fsm import MessageNotAccepted, State
from mediatumfsm.test import accumulatorfsm
from mediatumfsm.test.accumulatorfsm import Accumulator, AcceptInt, AcceptStr, Nothing, Int, Str
from mediatumfsm.test.regexfsm import RegexFSM

### fixtures

@fixture
def fsm():
    return Accumulator()

@fixture
def fsm_with_name():
    return Accumulator(name="test")

@fixture
def fsm_int():
    return Accumulator("Int", (23, None))

@fixture
def fsm_str():
    return Accumulator("Str", "spam")

@fixture
def fsm_regex():
    return RegexFSM()

### tests

def test_init(fsm):
    assert fsm.initial_data == False
    assert fsm.initial_state == Nothing


def test_can_access_states_in_module(fsm):
    assert issubclass(accumulatorfsm.Nothing, State)  # @UndefinedVariable
    assert issubclass(accumulatorfsm.Int, State)  # @UndefinedVariable
    assert issubclass(accumulatorfsm.Str, State)  # @UndefinedVariable

def test_init_with_values(fsm_int):
    assert fsm_int.data == (23, None)
    assert fsm_int.state == Int


def test_switch_to_str(fsm):
    state = fsm.send(AcceptStr())
    assert state == fsm.state == Str


def test_accept_str_value(fsm_str):
    fsm_str.send("|eggs")
    assert fsm_str.data == "spam|eggs"
    assert fsm_str.state == Str


def test_switch_to_int_from_str(fsm_str):
    fsm_str.send(AcceptInt())
    assert fsm_str.state == Int
    assert fsm_str.data == (0, None)


def test_handler_with_condition_int(fsm_int):
    fsm_int.send(23)
    assert fsm_int.data == (46, "odd")
    fsm_int.send(358)
    assert fsm_int.data == (404, "even")


def test_on_transition(fsm):
    accumulatorfsm.switched_to_int_times = 0
    fsm.send(AcceptInt())
    assert accumulatorfsm.switched_to_int_times == 1
    fsm.send(AcceptInt())
    assert accumulatorfsm.switched_to_int_times == 1


def test_regexfsm(fsm_regex):
    inputstr = "aaaaRSGTEgtfvok*fswgBBBGEGHTEG#"
    fsm_regex.parse(inputstr)
    assert fsm_regex.stopped

### failure tests

def test_wrong_message_type(fsm_int):
    with raises(MessageNotAccepted):
        fsm_int.send(dict())


def test_unhandled_case_wrong_value(fsm_int):
    with raises(ValueError):
        fsm_int.send("wrong string is wrong type")


def test_regexfsm_malformed_input(fsm_regex):
    malformed_inputs = [
         "aaRSGTEgtfvok*fswgBBBGEGHTEG#",
         "bla",
         "#",
         ]
    for inp in malformed_inputs:
        with raises(MessageNotAccepted):
            print("testing malformed string " + inp)
            fsm_regex.parse(inp)


def test_regexfsm_incomplete_input(fsm_regex):
    incomplete_inputs = [
         "",
         "a",
         "aaa",
         "aaaarghkrogkG",
         ]
    for inp in incomplete_inputs:
        print("testing malformed string " + inp)
        fsm_regex.parse(inp)
        assert not fsm_regex.stopped

### pydot

def test_pydot(fsm):
    fsm.create_graph(add_unhandled_case_handlers=True)
    result = fsm.dot()
    print(result)
    assert "digraph" in result
    assert "AcceptInt" in result
    assert "Accumulator" in result
    assert "m % 2 == 1" in result
    assert "even" in result
    assert "color=black" in result
    assert "color=green" in result
    assert "style=dashed" in result



# -*- coding: utf-8 -*-
import ast
from copy import deepcopy
from collections import Sequence
import inspect
import logging
import sys
import textwrap

from mediatumfsm._fsm import _StateMeta, _FSMMeta, _StateChange
from mediatumfsm.compat import itervalues, iteritems, iterkeys, with_metaclass
from itertools import chain


logg = logging.getLogger("fsm")

try:
    from pydot import Node, Edge, Dot
except ImportError:
    logg.warn("pydot not found. State Machine Diagrams won't work!")

### fsm exceptions

class FSMException(Exception):
    pass


class MessageNotAccepted(ValueError, FSMException):
    def __init__(self, rejected_message, current_state, allowed_messages):
        self.rejected_message = rejected_message
        self.allowed_messages = allowed_messages
        msg = "message {} {} in state {} not accepted. Accepted messages are: {}".format(
            rejected_message, rejected_message.__class__, current_state.name, allowed_messages)
        ValueError.__init__(self, msg)


class WrongResultType(ValueError, FSMException):
    def __init__(self, result, expected_result_type):
        msg = "wrong result {} : ({}), only instances of {} allowed!".format(
                                    result, result.__class__.__name__, expected_result_type)
        ValueError.__init__(self, msg)


def _match_handlers(handlers, message, data):
    try:
        handlers_for_msg_type = handlers[message.__class__]
    except KeyError:
        return None
    for h in handlers_for_msg_type:
        if h._condition is None or h._condition(message, data):
            logg.debug("handler %s matches", h)
            return h(message, data)

### state base class


class State(with_metaclass(_StateMeta)):
    """Subclass this to define your states. Must be an inner class in a FSM subclass.
    """
    @classmethod
    def set_action(cls, action):
        """Set callable which is used when FSM is in this state and fsm.action() is called.
        """
        cls.action = staticmethod(action)

    @staticmethod
    def action(*args, **kwargs):
        """Default: do nothing when fsm.action() is called
        """
        pass

    @classmethod
    def on_message(cls, message, data):
        """Called by FSM base class to handle messages in this state.
        """
        return _match_handlers(cls.handlers, message, data)


class End(State):
    """State which is reached when FSM is stopped.
    It's not possible to escape from this state.a
    """
    pass

### state changes

class Stay(_StateChange):
    pass

class Goto(_StateChange):
    def __init__(self, to):
        self.to = to

class Stop(_StateChange):
    def __init__(self, reason, data=None):
        self.reason = reason
        self.data = data

###  helpers for FSM


### FSM base

class FSM(with_metaclass(_FSMMeta)):
    """Subclass this to specify your FSM.
    The base FSM defines an End state which is inherited by all FSMs.
    Required:
      * override `initial_state_name` in your subclass to set the state which is used when a fresh FSM instance is created
    Optional:
      * override data_cls to restrict the FSM state data object to a certain type (or subclasses of the type)
    """
    initial_state_name = None
    initial_state = None # set automatically, do not override!
    data_cls = None # can be overriden
    End = End # default End state which is set when the FSM stops

    def __init__(self, state=None, data=None, name=None, fail_on_wrong_msg=True):
        """Initialize the state machine and set given state and data, if given.
        :param state: State to set FSM to. If None, use specified state given by `initial_state_name`.
        :param data: Initial state data. If None, use a copy of `initial_data`.
        :param name: Set a name for this FSM instance. Optional.
        :param fail_on_wrong_msg: True: Throw an exception if wrong messages (i.e. not accepted in current state) are sent to the FSM.
                                  False means that wrong msg are ignored.
        """
        self.fail_on_wrong_msg = fail_on_wrong_msg
        self.name = name
        self._graph = None
        if isinstance(state, str):
            try:
                state = self.states[state]
            except KeyError:
                raise ValueError("Unknown state {}, allowed states are {}".format(state, self.states))
        self.state = state if state else self.initial_state
        if data:
            self.data = data
        else:
            self.data = deepcopy(self.initial_data) if hasattr(self, "initial_data") else None


    def reset(self):
        """Reset the FSM to initial state and state data given in the FSM subclass."""
        self.state = self.initial_state
        self.data = deepcopy(self.initial_data) if hasattr(self, "initial_data") else None

    def _consume_handler_result(self, result):
        if not isinstance(result, Sequence):
            state_change = result
        elif len(result) == 2:
            state_change, new_data = result
            if new_data != self.data:
                if self.data_cls and new_data is not None and not isinstance(new_data, self.data_cls):
                    raise WrongResultType(new_data, self.data_cls)
                if logg.isEnabledFor(logging.DEBUG):
                    logg.debug("new data: %s", new_data)
                self.data = new_data
        else:
            raise NotImplementedError(result)
        if state_change == Stay or isinstance(state_change, Stay):
            logg.debug("no state change")
            return
        elif isinstance(state_change, Goto):
            logg.debug("transition from %s to %s", self.state, state_change.to)
            self.state = state_change.to
        elif state_change == Stop:
            logg.debug("finished")
            self.state = End
        elif isinstance(state_change, Stop):
            stop = state_change
            self.state = End
            logg.debug("finished, reason was %s, data %s", stop.reason, stop.data)
        else:
            raise NotImplementedError(state_change)
        return self.state

    def save_state(self):
        """Override this to specify how the current state and state data are saved in FSM.send() after message handling.
        See `SessionFSM` for an example how to save to a Flask-like Session object.
        """
        pass

    def send(self, message):
        """Send message to FSM and try to match handlers.
        First, all @case handlers in the current state are matched witch the given message and current state data.
        If no handler from the state matches, the @unhandled_case handlers specified for this FSM are tried.
        :param message: FSM message
        :throws MessageNotAccepted: Thrown if `self.fail_on_wrong_msg` is True and no handler matches.
        :throws WrongResultType: Thrown if `data_cls` is set and result type doesn't match `data_cls`.
        """
        logg.debug("got msg %s, current state %s with data %s", message, self.state.name, self.data)
        result = self.state.on_message(message, self.data)
        if result is None:
            # msg unhandled by state, try @unhandled_case common handlers
            result = _match_handlers(self.handlers, message, self.data)
            # still None?
            if result is None:
                if self.fail_on_wrong_msg:
                    raise MessageNotAccepted(message, self.state, self.accepted_messages)
                else:
                    return
        old_state = self.state
        new_state = self._consume_handler_result(result)
        if new_state:
            # run transition handlers
            for transition in ((State, State),
                               (State, new_state.__name__),
                               (old_state.__name__, State),
                               (old_state.__name__, new_state.__name__)):
                handler = self.transition_actions.get(transition)
                if handler:
                    handler(message, self.data, old_state, new_state)
        self.save_state()
        return new_state

    def _analyze_handler(self, h):
        # fix for the stupid code parser in inspect module...
        source, start = inspect.findsource(h)
        if "lambda" in source[start]:
            func_code = source[start] + "".join(inspect.getblock(source[start+1:]))
        else:
            func_code = "".join(inspect.getblock(source[start:]))
        code = textwrap.dedent(func_code)
        logg.debug("inspecting function\n %s", code)
        func_def = ast.parse(code).body[0]
        decorator_args = func_def.decorator_list[0].args
        # get condition from decorator if present, else None
        condition = None
        if len(decorator_args) > 1:
            cond = decorator_args[1]
            if isinstance(cond, ast.Name):
                condition = cond.id
            elif isinstance(cond, ast.Lambda):
                # extract lambda condition, ugly source hackery ;)
                condition = source[start].strip()[cond.body.col_offset:].rstrip("),")
        last_stmt = func_def.body[-1]
        if isinstance(last_stmt, ast.Return):
            return_value = last_stmt.value
        else:
            return None, condition
        if isinstance(return_value, ast.Tuple):
            state_change = return_value.elts[0]
        else:
            state_change = return_value
        expr = ast.Expression(state_change)
        ast.fix_missing_locations(expr)
        cc = compile(expr, "<expr_ast>", "eval")
#         ctx.update(self.states)
        mod = inspect.getmodule(self)
        try:
            change = eval(cc, mod.__dict__, self.states)
        except Exception as e:
            logg.error("failure analyzing handler '%s'. A programming error in your FSM maybe?", h.__name__, exc_info=1)
            logg.error("source was: \n%s", func_code)
            return None, condition
        return change, condition

    def create_graph(self, add_unhandled_case_handlers=False):
        """Create a graphviz graph (pydot) in memory for future graphviz code / image output.
        To render graphviz code or an image see the render_* methods.
        :param add_unhandled_case_handlers: If True, add all @unhandled_case handlers to the image
            (dotted arrows from each to every other state). Can be quite messy...
        """
        def _create_edge(change, src_node, label, style="solid", unhandled=False):
            style = "dashed" if unhandled else "solid"
            if change == Stay or isinstance(change, Stay):
                edge = Edge(src_node, src_node, color="blue", label=label, style=style)
            elif isinstance(change, Goto):
                dest_node = nodes[change.to]
                edge = Edge(src_node, dest_node, color="black", label=label, style=style)
            elif change == Stop or isinstance(change, Stop):
                stop = nodes[End]
                edge = Edge(src_node, stop, color="red", label=label, style=style)
            else:
                raise Exception("??! " + str(change))
            g.add_edge(edge)

        name = self.__class__.__name__
        g = Dot(graph_name=name,
                    labelloc="t", label="FSM: " + name, fontsize=18, fontcolor="blue")
        g.set_node_defaults(shape="ellipse", fontsize=12)
        g.set_edge_defaults(fontsize=13, labeldistance=3)
        nodes = {}
        handled_msg_types_per_state = {}
        for state in itervalues(self.states):
            if state == self.initial_state:
                color = "green"
            elif state == End:
                color = "red"
            else:
                color = "black"
            node = Node(state.name, color=color, style="rounded", shape="box")
            nodes[state] = node
            g.add_node(node)

        for src_state, src_node in iteritems(nodes):
            for msg_type, h_for_msg_type in iteritems(src_state.handlers):
                for h in h_for_msg_type:
                    change, condition = self._analyze_handler(h)
                    if condition:
                        label = " " + msg_type.__name__ + ": " + condition + " "
                    else:
                        label = " " + msg_type.__name__ + " "
                        # collect msg types which are finally handled by this state
                        handled_msg_types_per_state.setdefault(src_state, []).append(msg_type)
                    _create_edge(change, src_node, label)

        if add_unhandled_case_handlers:
            for msg_type, h_for_msg_type in iteritems(self.handlers):
                for h in h_for_msg_type:
                    change, condition = self._analyze_handler(h)
                    if change:
                        for src_state, src_node in iteritems(nodes):
                            if msg_type not in handled_msg_types_per_state.get(src_state, []):
                                _create_edge(change, src_node, msg_type.__name__, unhandled=True)
        self._graph = g

    @classmethod
    def generate_state_assignments(cls):
        TMPL = "{state} = {fsm}.{state}"
        lines = []
        for state_name in cls.states:
            lines.append(TMPL.format(fsm=cls.__name__, state=state_name))
        return "\n".join(lines)

    def render_dot(self, filepath):
        """Output graphviz code to `filepath`"""
        self._graph.write_raw(filepath)

    def render_svg(self, filepath):
        """Output svg code to `filepath`"""
        self._graph.write_svg(filepath)

    def render_png(self, filepath):
        """Output png image to `filepath`"""
        self._graph.write_png(filepath)

    def dot(self):
        """Get graphviz code."""
        return self._graph.to_string()

    @property
    def accepted_messages(self):
        """Gets messages which would be accepted in the current state (including @unhandled_case handlers).
        """
        return set(chain(iterkeys(self.state.handlers), iterkeys(self.handlers)))

    @property
    def accepted_messages_all_states(self):
        """Gets all messages which would be accepted in some state (or in a @unhandled_case handler).
        """
        accepted = []
        for state in itervalues(self.states):
            accepted.append(iterkeys(state.handlers))
        return set(chain(*accepted))

    @property
    def stopped(self):
        """Are we in the End state?"""
        return self.state is End

    def call(self, func, *args, **kwargs):
        """Call `func` with current state and state data as first arguments.
        *args and **kwargs are passed to `func`.
        """
        return func(self.state, self.data, *args, **kwargs)

    def action(self, *args, **kwargs):
        """Run action for current state with state and state data as first arguments.
        Results are returned. See `State.set_action()`"""
        return self.state.action(self.state, self.data, *args, **kwargs)


    def __eq__(self, other):
        return (self.name == other.name and
            self.state == other.state and
            self.data == other.data)


    def __hash__(self):
        return hash((self.name, self.state, self.name))

### state method decorators

def case(message_type, condition=None, pos=sys.maxsize):
    """
    Marks a static method in a State subclass as transition handler.
    The @staticmethod decorator doesn't have to be specified, this @case decorator handles that for you.
    :param message_type: Only messages with type `message_type` are matched by this handler.
    :param condition: Callable which is called with the message and current data as arguments.
        Returned truth value decides if handler matches.
    :param pos: Try to match this handler before every other handler with `pos`(other) > `pos`.
        If handlers have the same `pos`, the following rules apply:
        * handlers with condition are matched before handlers without
        * handlers which are defined earlier in the state class (lower line number) are matched first
    """

    def _case(f):
        f._message_type = message_type
        f._bound_to_state = True
        f._condition = condition
        f._pos = pos
        return staticmethod(f)
    return _case


def unhandled_case(message_type, condition=None, pos=sys.maxsize):
    """
    Marks a static method in a FSM subclass as transition handler for unhandled messages.
    These handlers are only tried if no handler in the current state matches.
    The @staticmethod decorator doesn't have to be specified, this @case decorator handles that for you.
    :param message_type: Only messages with type `message_type` are matched by this handler.
    :param condition: Callable which is called with the message and current data as arguments.
        Returned truth value decides if handler matches.
    :param pos: Try to match this handler before every other handler with `pos`(other) > `pos`.
        If handlers have the same `pos`, the following rules apply:
        * handlers with condition are matched before handlers without
        * handlers which are defined earlier in the state class (lower line number) are matched first
    """
    def _unhandled_case(f):
        f._message_type = message_type
        f._condition = condition
        f._bound_to_state = False
        f._pos = pos
        return staticmethod(f)
    return _unhandled_case


def on_transition(src_state, dest_state):
    """Marks a static method in a FSM subclass as transition action. These actions are called if the specified transition occurs.
    Returning Stay from a handler method doesn't call any actions.
    :param src_state: name of the State subclass before the transition or the State class itself to match all states
    :type src_state: str
    :param dest_state: Same as src_state for the new state. "End" can be given to match transitions to the End state (i.e. handler returned Stop)
    """
    def _on_transition(f):
        f._src_state = src_state
        f._dest_state = dest_state
        return staticmethod(f)
    return _on_transition